Enum orx_imp_vec::prelude::PinnedVecGrowthError
source ยท pub enum PinnedVecGrowthError {
CanOnlyGrowWhenVecIsAtCapacity,
FailedToGrowWhileKeepingElementsPinned,
}Expand description
Error occurred during an attempt to increase capacity of the pinned vector.
Variants
CanOnlyGrowWhenVecIsAtCapacity
An error stating that the vector is only allowed to grow if its entire current capacity is used.
FailedToGrowWhileKeepingElementsPinned
An error which is observed when a pinned vector attempted to increase its capacity while keeping its already added elements pinned in their locations.