Enum orx_imp_vec::prelude::SplitVecSlice

source ·
pub enum SplitVecSlice<'a, T> {
    Ok(&'a [T]),
    Fragmented(usize, usize),
    OutOfBounds,
}
Returns the result of trying to get a slice as a contagious memory from the split vector.

Variants



Ok(&'a [T])

The desired range completely belongs to one fragment and the slice can be provided.



Fragmented(usize, usize)

The desired range is split to at least two fragments. The tuple contains indices of the fragments containing the first and last element of the desired range.



OutOfBounds

An error case where the desired range is out of bounds of the vector.